import jwt
import json as json_lib
import time

from typing import Any

from vantage6.common.client.client_base import ClientBase
from vantage6.common import base64s_to_bytes, bytes_to_base64s
from vantage6.common.task_status import TaskStatus
from vantage6.tools.serialization import serialize
from vantage6.tools.util import info


class AlgorithmClient(ClientBase):
    """
    Interface to communicate between the algorithm container and the central
    server via a local proxy server.

    An algorithm container cannot communicate directly to the
    central server as it has no internet connection. The algorithm can,
    however, talk to a local proxy server which has interface to the central
    server. This way we make sure that the algorithm container does not share
    details with others, and we also can encrypt the results for a specific
    receiver. Thus, this not a interface to the central server but to
    the local proxy server - however, the interface looks identical to make
    it easier to use.

    Parameters
    ----------
    token: str
        JWT (container) token, generated by the node the algorithm container
        runs on
    *args, **kwargs
        Arguments passed to the parent ClientBase class.
    """

    def __init__(self, token: str, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # obtain the identity from the token
        jwt_payload = jwt.decode(
            token, options={"verify_signature": False})

        container_identity = jwt_payload['sub']

        self.image = container_identity.get("image")
        self.databases = container_identity.get('databases', [])
        self.node_id = container_identity.get("node_id")
        self.collaboration_id = container_identity.get("collaboration_id")
        self.organization_id = container_identity.get("organization_id")
        self.log.info(
            f"Container in collaboration_id={self.collaboration_id} \n"
            f"Key created by node_id {self.node_id} \n"
            f"Can only use image={self.image}"
        )

        # attach sub-clients
        self.run = self.Run(self)
        self.result = self.Result(self)
        self.task = self.Task(self)
        self.vpn = self.VPN(self)
        self.organization = self.Organization(self)
        self.collaboration = self.Collaboration(self)
        self.node = self.Node(self)

        self._access_token = token

    def request(self, *args, **kwargs) -> dict:
        """
        Make a request to the central server. This overwrites the parent
        function so that containers will not try to refresh their token, which
        they would be unable to do.

        Parameters
        ----------
        *args, **kwargs
            Arguments passed to the parent ClientBase.request function.

        Returns
        -------
        dict
            Response from the central server.
        """
        return super().request(*args, **kwargs, retry=False)

    def wait_for_results(self, task_id: int, interval: float = 1) -> list:
        """
        Poll the central server until results are available and then return
        them.

        Parameters
        ----------
        task_id: int
            ID of the task for which the results should be obtained.
        interval: float
            Interval in seconds to wait between checking server for results.

        Returns
        -------
        list
            List of task results.
        """
        while self.task.get(task_id)['status'] != TaskStatus.COMPLETED:
            info(f"Waiting for results of task {task_id}...")
            time.sleep(interval)
        info("Done!")

        return self.result.from_task(task_id)

    def _multi_page_request(self, endpoint: str, params: dict = None) -> dict:
        """
        Make multiple requests to the central server to get all pages of a list
        of results.

        Parameters
        ----------
        endpoint: str
            Endpoint to which the request should be made.
        params: dict
            Parameters to be passed to the request.

        Returns
        -------
        dict
            Response from the central server.
        """
        if params is None:
            params = {}
        # get first page
        page = 1
        params["page"] = page
        response = self.request(endpoint, params=params)

        # append next pages (if any)
        links = response.get("links")
        while links and links.get("next"):
            page += 1
            params["page"] = page
            response["data"] += self.request(endpoint, params=params)["data"]
            links = response.get("links")

        return response['data']

    class Run(ClientBase.SubClient):
        """
        Algorithm Run client for the algorithm container.

        This client is used to obtain algorithm runs of tasks with the same
        job_id from the central server.
        """

        def get(self, id_) -> dict:
            """
            Obtain a specific algorithm run from the central server.

            Parameters
            ----------
            id_: int
                ID of the algorithm run that should be obtained.

            Returns
            -------
            dict
                Algorithm run data.
            """
            return self.parent.request(f"run/{id_}")

        def from_task(self, task_id: int) -> list:
            """
            Obtain algorithm runs from a specific task at the server.

            Containers are allowed to obtain the runs of their children
            (having the same job_id at the server). The permissions are checked
            at te central server.

            Note that the returned results are not decrypted. The algorithm is
            responsible for decrypting the results.

            Parameters
            ----------
            task_id: int
                ID of the task from which you want to obtain the algorithm runs

            Returns
            -------
            list
                List of algorithm run data. The type of the results depends on
                the algorithm.
            """
            # TODO do we need this function? It may be used to collect data
            # on subtasks but usually only the results are accessed, which is
            # done with the function below.
            return self.parent._multi_page_request(
                "run", params={"task_id": task_id}
            )

    class Result(ClientBase.SubClient):
        """
        Result client for the algorithm container.

        This client is used to get results from the central server.
        """
        def get(self, id_: int) -> Any:
            """
            Obtain a specific result from the central server.

            Parameters
            ----------
            id_: int
                ID of the algorithm run of which the result should be obtained.

            Returns
            -------
            Any
                Result of the algorithm run.
            """
            response = self.parent.request(f"result/{id_}")

            # Encryption is not done at the client level for the container. The
            # algorithm developer is responsible for decrypting the results.
            self.parent.log.info('--> Attempting to decode results!')
            result = None
            if response.get('result'):
                result = json_lib.loads(
                    base64s_to_bytes(response.get('result')).decode()
                )
            return result

        def from_task(self, task_id: int) -> list[Any]:
            """
            Obtain results from a specific task at the server.

            Containers are allowed to obtain the results of their children
            (having the same job_id at the server). The permissions are checked
            at te central server.

            Results are decrypted by the proxy server and decoded here before
            returning them to the algorithm.

            Parameters
            ----------
            task_id: int
                ID of the task from which you want to obtain the results

            Returns
            -------
            list[Any]
                List of results. The type of the results depends on the
                algorithm.
            """
            results = self.parent._multi_page_request(
                "result", params={"task_id": task_id}
            )

            # Encryption is not done at the client level for the container. The
            # algorithm developer is responsible for decrypting the results.
            decoded_results = []
            try:
                decoded_results = [
                    json_lib.loads(
                        base64s_to_bytes(result.get("result")).decode()
                    )
                    for result in results if result.get("result")
                ]
            except Exception as e:
                self.parent.log.error('Unable to load results')
                self.parent.log.debug(e)

            return decoded_results

    class Task(ClientBase.SubClient):
        """
        A task client for the algorithm container.

        It provides functions to get task information and create new tasks.
        """
        def get(self, task_id: int) -> dict:
            """
            Retrieve a task at the central server.

            Parameters
            ----------
            task_id : int
                ID of the task to retrieve

            Returns
            -------
            dict
                Dictionary containing the task information
            """
            return self.parent.request(
                f"task/{task_id}"
            )

        def create(
            self, input_: bytes, organizations: list[int] = None,
            name: str = "subtask", description: str = None
        ) -> dict:
            """
            Create a new (child) task at the central server.

            Containers are allowed to create child tasks (having the
            same job_id) at the central server. The docker image must
            be the same as the docker image of this container self.

            Parameters
            ----------
            input_ : bytes
                Input to the task. Should be b64 encoded.
            organizations : list[int]
                List of organization IDs that should execute the task.
            name: str, optional
                Name of the subtask
            description : str, optional
                Description of the subtask

            Returns
            -------
            dict
                Dictionary containing information on the created task
            """
            if not organizations:
                organizations = []
            self.parent.log.debug(f"Creating new subtask for {organizations}")

            description = (
                description or
                f"task from container on node_id={self.parent.node_id}"
            )

            # serializing input. Note that the input is not encrypted here, but
            # in the proxy server (self.parent.request())
            serialized_input = bytes_to_base64s(serialize(input_))
            organization_json_list = []
            for org_id in organizations:
                organization_json_list.append(
                    {
                        "id": org_id,
                        "input": serialized_input
                    }
                )

            return self.parent.request('task', method='post', json={
                "name": name,
                "image": self.parent.image,
                "collaboration_id": self.parent.collaboration_id,
                "description": description,
                "organizations": organization_json_list,
                "databases": self.parent.databases
            })

    class VPN(ClientBase.SubClient):
        """
        A VPN client for the algorithm container.

        It provides functions to obtain the IP addresses of other containers.
        """
        def get_addresses(
            self, only_children: bool = False, only_parent: bool = False,
            include_children: bool = False, include_parent: bool = False,
            label: str = None
        ) -> list[dict] | dict:
            """
            Get information about the VPN IP addresses and ports of other
            algorithm containers involved in the current task. These addresses
            can be used to send VPN communication to.

            Parameters
            ----------
            only_children : bool, optional
                Only return the IP addresses of the children of the current
                task, by default False. Incompatible with only_parent.
            only_parent : bool, optional
                Only return the IP address of the parent of the current task,
                by default False. Incompatible with only_children.
            include_children : bool, optional
                Include the IP addresses of the children of the current task,
                by default False. Incompatible with only_parent, superseded
                by only_children.
            include_parent : bool, optional
                Include the IP address of the parent of the current task, by
                default False. Incompatible with only_children, superseded by
                only_parent.
            label : str, optional
                The label of the port you are interested in, which is set
                in the algorithm Dockerfile. If this parameter is set, only
                the ports with this label will be returned.

            Returns
            -------
            list[dict] | dict
                List of dictionaries containing the IP address and port number,
                and other information to identify the containers. If obtaining
                the VPN addresses from the server fails, a dictionary with a
                'message' key is returned instead.
            """
            results = self.parent.request("vpn/algorithm/addresses", params={
                "only_children": only_children,
                "only_parent": only_parent,
                "include_children": include_children,
                "include_parent": include_parent,
                "label": label
            })

            if 'addresses' not in results:
                return {'message': 'Obtaining VPN addresses failed!'}

            return results['addresses']

        def get_parent_address(self) -> dict:
            """
            Get the IP address and port number of the parent of the current
            task.

            Returns
            -------
            dict
                Dictionary containing the IP address and port number, and other
                information to identify the containers. If obtaining the VPN
                addresses from the server fails, a dictionary with a 'message'
                key is returned instead.
            """
            return self.get_addresses(only_parent=True)

        def get_child_addresses(self) -> list[dict]:
            """
            Get the IP addresses and port numbers of the children of the
            current task.

            Returns
            -------
            List[dict]
                List of dictionaries containing the IP address and port number,
                and other information to identify the containers. If obtaining
                the VPN addresses from the server fails, a dictionary with a
                'message' key is returned instead.
            """
            return self.get_addresses(only_children=True)

    class Organization(ClientBase.SubClient):
        """
        Get information about organizations in the collaboration.
        """
        def get(self, id_: int) -> dict:
            """
            Get an organization by ID.

            Parameters
            ----------
            id: int
                ID of the organization to retrieve

            Returns
            -------
            dict
                Dictionary containing the organization data.
            """
            return self.parent.request(f"organization/{id_}")

        def list(self) -> list[dict]:
            """
            Obtain all organization in the collaboration.

            The container runs in a Node which is part of a single
            collaboration. This method retrieves all organization data that are
            within that collaboration. This can be used to target specific
            organizations in a collaboration.

            Returns
            -------
            list[dict]
                List of organizations in the collaboration.
            """
            return self.parent._multi_page_request(
                endpoint="organization", params={
                    "collaboration_id": self.parent.collaboration_id
                }
            )

    class Collaboration(ClientBase.SubClient):
        """
        Get information about the collaboration.
        """
        def get(self) -> dict:
            """
            Get the collaboration data.

            Returns
            -------
            dict
                Dictionary containing the collaboration data.
            """
            return self.parent.request(
                f"collaboration/{self.parent.collaboration_id}")

    class Node(ClientBase.SubClient):
        """
        Get information about the node.
        """
        def get(self) -> dict:
            """
            Get the node data.

            Returns
            -------
            dict
                Dictionary containing data on the node this algorithm is
                running on.
            """
            return self.parent.request(f"node/{self.parent.node_id}")



import pandas as pd
import json
import logging

from typing import Any
from importlib import import_module
from copy import deepcopy

from vantage6.tools.wrappers import select_wrapper
from vantage6.tools.util import info

module_name = __name__.split('.')[1]


class MockAlgorithmClient:
    """
    The MockAlgorithmClient mimics the behaviour of the AlgorithmClient. It
    can be used to mock the behaviour of the AlgorithmClient and its
    communication with the server.

    Parameters
    ----------
    datasets : list[list[dict]]
        A list that contains the datasets that are used in the mocked
        algorithm. The inner list contains the datasets for each organization;
        the outer list is for each organization. A single dataset should be
        described as a dictionary with the same keys as in a node
        configuration:

        - database: str (path to file or SQL connection string) or pd.DataFrame
        - type (str, e.g. "csv" or "sql")
        - input_data (dict). The input data that is normally passed to the
          algorithm wrapper, e.g. a worksheet for an excel database.

        Note that if the database is a pandas DataFrame, the type and
        input_data keys are not required.
    module : str
        The name of the module that contains the algorithm.
    collaboration_id : int, optional
        Sets the mocked collaboration id to this value. Defaults to 1.
    organization_ids : list[int], optional
        Set the organization ids to this value. The first value is used for
        this organization, the rest for child tasks. Defaults to [0, 1, 2, ..].
    node_ids: list[int], optional
        Set the node ids to this value. The first value is used for this node,
        the rest for child tasks. Defaults to [0, 1, 2, ...].
    """
    def __init__(
        self, datasets: list[dict], module: str, collaboration_id: int = None,
        organization_ids: int = None, node_ids: int = None,
    ) -> None:
        self.log = logging.getLogger(module_name)
        self.n = len(datasets)

        if organization_ids and len(organization_ids) == self.n:
            self.all_organization_ids = organization_ids
        else:
            default_organization_ids = list(range(self.n))
            if organization_ids:
                self.log.warning(
                    "The number of organization ids (%s) does not match the "
                    "number of datasets (#=%s), using default values (%s) "
                    "instead",
                    organization_ids,
                    self.n,
                    default_organization_ids,
                )
            self.all_organization_ids = default_organization_ids

        self.organization_id = self.all_organization_ids[0]

        if node_ids and len(node_ids) == self.n:
            self.all_node_ids = node_ids
        else:
            default_node_ids = list(range(self.n))
            if node_ids:
                self.log.warning(
                    "The number of node ids (%s) does not match the number of "
                    "datasets (#=%s), using default values (%s) instead",
                    node_ids,
                    self.n,
                    default_node_ids,
                )
            self.all_node_ids = default_node_ids
        self.node_id = self.all_node_ids[0]

        self.datasets_per_org = {}
        self.organizations_with_data = []
        for idx, org_datasets in enumerate(datasets):
            org_id = self.all_organization_ids[idx]
            self.organizations_with_data.append(org_id)
            org_data = []
            for dataset in org_datasets:
                if isinstance(dataset["database"], pd.DataFrame):
                    org_data.append(dataset["database"])
                else:
                    wrapper = select_wrapper(dataset["type"])
                    org_data.append(
                        wrapper.load_data(
                            dataset["database"],
                            dataset["input_data"] if "input_data" in dataset
                            else {}
                        )
                    )
            self.datasets_per_org[org_id] = org_data

        self.collaboration_id = collaboration_id if collaboration_id else 1
        self.module_name = module
        self.tasks = []
        self.runs = []
        self.results = []

        self.image = 'mock_image'
        self.database = 'mock_database'

        self.task = self.Task(self)
        self.result = self.Result(self)
        self.organization = self.Organization(self)
        self.collaboration = self.Collaboration(self)
        self.node = self.Node(self)

    def wait_for_results(self, task_id: int, *args, **kwargs) -> list:
        """
        Mock waiting for results - just return the results as tasks are
        completed synchronously in the mock client.

        Parameters
        ----------
        task_id: int
            ID of the task for which the results should be obtained.

        Returns
        -------
        list
            List of task results.
        """
        info('Mocking waiting for results')
        return self.result.from_task(task_id)

    class SubClient:
        """
        Create sub groups of commands using this SubClient

        Parameters
        ----------
        parent : MockAlgorithmClient
            The parent client
        """
        def __init__(self, parent) -> None:
            self.parent: MockAlgorithmClient = parent

    class Task(SubClient):
        """
        Task subclient for the MockAlgorithmClient
        """
        def __init__(self, parent) -> None:
            super().__init__(parent)
            self.last_result_id = 0

        def create(
            self, input_: dict, organizations: list[int],
            name: str = "mock", description: str = "mock", *args, **kwargs
        ) -> int:
            """
            Create a new task with the MockProtocol and return the task id.

            Parameters
            ----------
            input_ : dict
                The input data that is passed to the algorithm. This should at
                least  contain the key 'method' which is the name of the method
                that should be called. Another often used key is 'master' which
                indicates that this container is a master container. Other keys
                depend on the algorithm.
            organizations : list[int]
                A list of organization ids that should run the algorithm.
            name : str, optional
                The name of the task, by default "mock"
            description : str, optional
                The description of the task, by default "mock"

            Returns
            -------
            int
                The id of the task.
            """
            if not len(organizations):
                raise ValueError(
                    "No organization ids provided. Cannot create a task for "
                    "zero organizations."
                )

            module = import_module(self.parent.module_name)

            # extract method from lib and input
            method_name = input_.get("method")
            method = getattr(module, method_name)

            # get input
            args = input_.get("args", [])
            kwargs = input_.get("kwargs", {})

            # get data for organization
            for org_id in organizations:
                # When creating a child task, pass the parent's datasets and
                # client to the child. By passing also the client, the child
                # has access to the same IDs specified
                data = self.parent.datasets_per_org[org_id]
                client_copy = deepcopy(self.parent)
                client_copy.node_id = self._select_node(org_id)
                client_copy.organization_id = org_id
                result = method(
                    mock_client=self.parent,
                    mock_data=data,
                    *args, **kwargs
                )

                self.last_result_id += 1
                self.parent.results.append({
                    "id": self.last_result_id,
                    "result": json.dumps(result),
                    "run": {
                        "id": self.last_result_id,
                        "link": f"/api/run/{self.last_result_id}",
                        "methods": ["GET", "PATCH"]
                    },
                })
                self.parent.runs.append({
                    "id": self.last_result_id,
                    "started_at": "2021-01-01T00:00:00.000000",
                    "assigned_at": "2021-01-01T00:00:00.000000",
                    "finished_at": "2021-01-01T00:00:00.000000",
                    "log": "mock_log",
                    "ports": [],
                    "status": "completed",
                    "input": json.dumps(input_),
                    "results": {
                        "id": self.last_result_id,
                        "link": f"/api/result/{self.last_result_id}",
                        "methods": ["GET", "PATCH"]
                    },
                    "node": {
                        "id": org_id,
                        "ip": None,
                        "name": "mock_node",
                        "status": "online",
                    },
                    "organization": {
                        "id": org_id,
                        "link": f"/api/organization/{org_id}",
                        "methods": ["GET", "PATCH"]
                    },
                    "task": {
                        "id": len(self.parent.tasks),
                        "link": f"/api/task/{len(self.parent.tasks)}",
                        "methods": ["GET", "PATCH"]
                    },
                })

            id_ = len(self.parent.tasks)
            collab_id = self.parent.collaboration_id
            task = {
                "id": id_,
                "runs": f"/api/run?task_id={id_}",
                "status": "completed",
                "name": name,
                "databases": ["mock"],
                "description": description,
                "image": "mock_image",
                "init_user": {
                    "id": 1,
                    "link": "/api/user/1",
                    "methods": ["GET", "DELETE", "PATCH"]
                },
                "init_org": {
                    "id": self.parent.organization_id,
                    "link": f"/api/organization/{self.parent.organization_id}",
                    "methods": ["GET", "PATCH"]
                },
                "parent": None,
                "collaboration": {
                    "id": collab_id,
                    "link": f"/api/collaboration/{collab_id}",
                    "methods": ["DELETE", "PATCH", "GET"]
                },
                "job_id": 1,
                "children": None,
            }
            self.parent.tasks.append(task)
            return task

        def get(self, task_id: int) -> dict:
            """
            Return the task with the given id.

            Parameters
            ----------
            task_id : int
                The id of the task.

            Returns
            -------
            dict
                The task details.
            """
            if task_id >= len(self.parent.tasks):
                return {
                    "msg": f"Could not find task with id {task_id}"
                }
            return self.parent.tasks[task_id]

        def _select_node(self, org_id: int) -> int:
            """
            Select a node for the given organization id.

            Parameters
            ----------
            org_id : int
                The organization id.

            Returns
            -------
            int
                The node id.
            """
            if not self.parent.all_node_ids or \
                    not self.parent.all_organization_ids or \
                    org_id not in self.parent.all_organization_ids:
                return org_id
            org_idx = self.parent.all_organization_ids.index(org_id)
            return self.parent.all_node_ids[org_idx]

    class Run(SubClient):
        """
        Run subclient for the MockAlgorithmClient
        """
        def get(self, id_: int) -> dict:
            """
            Get mocked run by ID

            Parameters
            ----------
            id_ : int
                The id of the run.

            Returns
            -------
            dict
                A mocked run.
            """
            for run in self.parent.runs:
                if run.get("id") == id_:
                    return run
            return {
                "msg": f"Could not find run with id {id_}"
            }

        def from_task(self, task_id: int) -> list[dict]:
            """
            Get mocked runs by task ID

            Parameters
            ----------
            task_id : int
                The id of the task.

            Returns
            -------
            list[dict]
                A list of mocked runs.
            """
            runs = []
            for run in self.parent.runs:
                if run.get("task").get("id") == task_id:
                    runs.append(run)
            return runs

    class Result(SubClient):
        """
        Result subclient for the MockAlgorithmClient
        """
        def get(self, id_: int) -> Any:
            """
            Get mocked result by ID

            Parameters
            ----------
            id_ : int
                The id of the result.

            Returns
            -------
            Any
                A mocked result.
            """
            for result in self.parent.results:
                if result.get("id") == id_:
                    return json.loads(result.get("result"))
            return {
                "msg": f"Could not find result with id {id_}"
            }

        def from_task(self, task_id: int) -> list[Any]:
            """
            Return the results of the task with the given id.

            Parameters
            ----------
            task_id : int
                The id of the task.

            Returns
            -------
            list[Any]
                The results of the task.
            """
            results = []
            for task in self.parent.tasks:
                if task.get("id") == task_id:
                    for result in task.get("results"):
                        results.append(json.loads(result.get("result")))
            return results

    class Organization(SubClient):
        """
        Organization subclient for the MockAlgorithmClient
        """
        def get(self, id_) -> dict:
            """
            Get mocked organization by ID

            Parameters
            ----------
            id_ : int
                The id of the organization.

            Returns
            -------
            dict
                A mocked organization.
            """
            if not id_ == self.parent.organization_id and \
                    id_ not in self.parent.organizations_with_data:
                return {
                    "msg": f"Organization {id_} not found."
                }
            return {
                "id": id_,
                "name": f"mock-{id_}",
                "domain": f"mock-{id_}.org",
                "address1": "mock",
                "address2": "mock",
                "zipcode": "mock",
                "country": "mock",
                "public_key": "mock",
                "collaborations": f"/api/collaboration?organization_id={id_}",
                "users": f"/api/user?organization_id={id_}",
                "tasks": f"/api/task?init_org_id={id_}",
                "nodes": f"/api/node?organization_id={id_}",
                "runs": f"/api/run?organization_id={id_}"
            }

        def list(self) -> list[dict]:
            """
            Get mocked organizations in the collaboration.

            Returns
            -------
            list[dict]
                A list of mocked organizations in the collaboration.
            """
            organizations = []
            for i in self.parent.all_organization_ids:
                organizations.append(self.get(i))
            return organizations

    class Collaboration(SubClient):
        """
        Collaboration subclient for the MockAlgorithmClient
        """
        def get(self, is_encrypted: bool = True) -> dict:
            """
            Get mocked collaboration

            Parameters
            ----------
            is_encrypted : bool
                Whether the collaboration is encrypted or not. Default True.

            Returns
            -------
            dict
                A mocked collaboration.
            """
            collab_id = self.parent.collaboration_id
            return {
                "id": collab_id,
                "name": "mock-collaboration",
                "encrypted": is_encrypted,
                "tasks": f"/api/task?collaboration_id={collab_id}",
                "nodes": f"/api/node?collaboration_id={collab_id}",
                "organizations":
                    f"/api/organization?collaboration_id={collab_id}"
            }

    class Node(SubClient):
        """
        Node subclient for the MockAlgorithmClient
        """
        def get(self, is_online: bool = True) -> dict:
            """
            Get mocked node

            Parameters
            ----------
            is_online : bool
                Whether the node is online or not. Default True.

            Returns
            -------
            dict
                A mocked node.
            """
            node_id = self.parent.node_id
            collab_id = self.parent.collaboration_id
            return {
                "id": node_id,
                "name": "mock-node",
                "status": "online" if is_online else "offline",
                "ip": "1.2.3.4",
                "config": {
                    "key": "value",
                },
                "collaboration": {
                    "id": collab_id,
                    "link": f"/api/collaboration/{collab_id}",
                    "methods": [
                        "DELETE",
                        "PATCH",
                        "GET"
                    ]
                },
                "last_seen": "2021-01-01T00:00:00.000000",
                "type": "node",
                "organization": {
                    "id": node_id,
                    "link": f"/api/organization/{node_id}",
                    "methods": [
                        "GET",
                        "PATCH"
                    ]
                },
            }

    # TODO implement the get_addresses method before using this part
    # class VPN(SubClient):
    #     """
    #     VPN subclient for the MockAlgorithmClient
    #     """
    #     def get_addresses(
    #         self, only_children: bool = False, only_parent: bool = False,
    #         include_children: bool = False, include_parent: bool = False,
    #         label: str = None
    #     ) -> list[dict] | dict:
    #         """
    #         Mock VPN IP addresses and ports of other algorithm containers in
    #         the current task.

    #         Parameters
    #         ----------
    #         only_children : bool, optional
    #             Only return the IP addresses of the children of the current
    #             task, by default False. Incompatible with only_parent.
    #         only_parent : bool, optional
    #             Only return the IP address of the parent of the current task,
    #             by default False. Incompatible with only_children.
    #         include_children : bool, optional
    #             Include the IP addresses of the children of the current task,
    #             by default False. Incompatible with only_parent, superseded
    #             by only_children.
    #         include_parent : bool, optional
    #             Include the IP address of the parent of the current task, by
    #             default False. Incompatible with only_children, superseded by
    #             only_parent.
    #         label : str, optional
    #             The label of the port you are interested in, which is set
    #             in the algorithm Dockerfile. If this parameter is set, only
    #             the ports with this label will be returned.

    #         Returns
    #         -------
    #         list[dict] | dict
    #             List of dictionaries containing the IP address and port number,
    #             and other information to identify the containers. If obtaining
    #             the VPN addresses from the server fails, a dictionary with a
    #             'message' key is returned instead.
    #         """
    #         pass

    #     def get_parent_address(self) -> dict:
    #         """
    #         Get the IP address and port number of the parent of the current
    #         task.

    #         Returns
    #         -------
    #         dict
    #             Dictionary containing the IP address and port number, and other
    #             information to identify the containers. If obtaining the VPN
    #             addresses from the server fails, a dictionary with a 'message'
    #             key is returned instead.
    #         """
    #         return self.get_addresses(only_parent=True)

    #     def get_child_addresses(self) -> list[dict]:
    #         """
    #         Get the IP addresses and port numbers of the children of the
    #         current task.

    #         Returns
    #         -------
    #         List[dict]
    #             List of dictionaries containing the IP address and port number,
    #             and other information to identify the containers. If obtaining
    #             the VPN addresses from the server fails, a dictionary with a
    #             'message' key is returned instead.
    #         """
    #         return self.get_addresses(only_children=True)



from pytest import mark

from vantage6.tools import serialization
import pandas as pd


@mark.parametrize("data,target", [
    # Default serialization
    ([1, 2, 3], '[1, 2, 3]'),
    ('hello', '"hello"'),
    ({'hello': 'goodbye'}, '{"hello": "goodbye"}'),

    # Pandas serialization
    (pd.DataFrame([[1, 2, 3]], columns=['one', 'two', 'three']),
     '{"one":{"0":1},"two":{"0":2},"three":{"0":3}}'),
    (pd.Series([1, 2, 3]), '{"0":1,"1":2,"2":3}')
])
def test_json_serialization(data, target):
    result = serialization.serialize(data)

    assert target == result.decode()



from pathlib import Path
from vantage6.common.client import deserialization

SIMPLE_TARGET_DATA = {'key': 'value'}


def test_deserialize_json(tmp_path: Path):
    data = '{"key": "value"}'
    json_path = tmp_path / 'jsonfile.json'
    json_path.write_text(data)

    with json_path.open('r') as f:
        result = deserialization.deserialize(f)

        assert SIMPLE_TARGET_DATA == result



"""
This module contains algorithm wrappers. These wrappers are used to provide
different data adapters to the algorithms. This way we ony need to write the
algorithm once and can use it with different data adapters.

Currently the following wrappers are available:
    - ``DockerWrapper`` (= ``CSVWrapper``)
    - ``SparqlDockerWrapper``
    - ``ParquetWrapper``
    - ``SQLWrapper``
    - ``OMOPWrapper``
    - ``ExcelWrapper``

When writing the Docker file for the algorithm, the correct wrapper will
automatically be selected based on the database type. The database type is set
by the vantage6 node based on its configuration file.
"""
from __future__ import annotations
import io
import pandas

from abc import ABC, abstractmethod
from SPARQLWrapper import SPARQLWrapper, CSV

from vantage6.tools.util import info, error

_SPARQL_RETURN_FORMAT = CSV


def select_wrapper(database_type: str) -> WrapperBase | None:
    """
    Select the correct wrapper based on the database type.

    Parameters
    ----------
    database_type : str
        The database type to select the wrapper for.

    Returns
    -------
    derivative of WrapperBase | None
        The wrapper for the specified database type. None if the database type
        is not supported by a wrapper.
    """
    if database_type == "csv":
        return CSVWrapper()
    elif database_type == "excel":
        return ExcelWrapper()
    elif database_type == "sparql":
        return SparqlDockerWrapper()
    elif database_type == "parquet":
        return ParquetWrapper()
    elif database_type == "sql":
        return SQLWrapper()
    elif database_type == "omop":
        return OMOPWrapper()
    else:
        return None


class WrapperBase(ABC):
    @staticmethod
    @abstractmethod
    def load_data(database_uri: str, input_data: dict):
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the database to read
        input_data : dict
            User defined input, which may contain a query for the database
        """
        pass


class CSVWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the csv file, supplied by te node
        input_data : dict
            Unused, as csv files do not require a query

        Returns
        -------
        pandas.DataFrame
            The data from the csv file
        """
        return pandas.read_csv(database_uri)


class ExcelWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the excel file, supplied by te node
        input_data : dict
            May contain a 'sheet_name', which is passed to pandas.read_excel

        Returns
        -------
        pandas.DataFrame
            The data from the excel file
        """
        # The default sheet_name is 0, which is the first sheet
        sheet_name = input_data.get('sheet_name', 0)
        if sheet_name:
            info(f"Reading sheet '{sheet_name}' from excel file")
        return pandas.read_excel(database_uri, sheet_name=sheet_name)


class SparqlDockerWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the triplestore, supplied by te node
        input_data : dict
            Can contain a 'query', to retrieve the data from the triplestore

        Returns
        -------
        pandas.DataFrame
            The data from the triplestore
        """
        if 'query' not in input_data:
            error("No query in the input specified. Exiting ...")
        query = input_data['query']
        return SparqlDockerWrapper._query_triplestore(database_uri, query)

    @staticmethod
    def _query_triplestore(endpoint: str, query: str) -> pandas.DataFrame:
        """
        Send a query to a triplestore and return the result as a pandas
        DataFrame.

        Parameters
        ----------
        endpoint : str
            URI of the triplestore
        query : str
            The query to send to the triplestore

        Returns
        -------
        pandas.DataFrame
            The result of the query
        """
        sparql = SPARQLWrapper(endpoint, returnFormat=_SPARQL_RETURN_FORMAT)
        sparql.setQuery(query)

        result = sparql.query().convert().decode()

        return pandas.read_csv(io.StringIO(result))


class ParquetWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the parquet file, supplied by te node
        input_data : dict
            Unused, as no additional settings are required

        Returns
        -------
        pandas.DataFrame
            The data from the parquet file
        """
        return pandas.read_parquet(database_uri)


class SQLWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the sql database, supplied by te node
        input_data : dict
            Contain a 'query', to retrieve the data from the database

        Returns
        -------
        pandas.DataFrame
            The data from the database
        """
        if 'query' not in input_data:
            error("No query in the input specified. Exiting ...")
        return pandas.read_sql(database_uri, input_data['query'])


class OMOPWrapper(WrapperBase):
    @staticmethod
    def load_data(database_uri: str, input_data: dict) -> pandas.DataFrame:
        """
        Load the local privacy-sensitive data from the database.

        Parameters
        ----------
        database_uri : str
            URI of the OMOP database, supplied by te node
        input_data : dict
            Contain a JSON cohort definition from the ATLAS tool, to retrieve
            the data from the database

        Returns
        -------
        pandas.DataFrame
            The data from the database
        """
        # TODO: parse the OMOP json and convert to SQL
        return pandas.read_sql(database_uri, input_data['query'])



import os
import importlib
import traceback

from typing import Any

from vantage6.common.client import deserialization
from vantage6.tools import serialization
from vantage6.tools.util import info, error
from vantage6.tools.exceptions import DeserializationException


def wrap_algorithm(module: str, log_traceback: bool = True) -> None:
    """
    Wrap an algorithm module to provide input and output handling for the
    vantage6 infrastructure.

    Data is received in the form of files, whose location should be
    specified in the following environment variables:

    - ``INPUT_FILE``: input arguments for the algorithm. This file should be
      encoded in JSON format.
    - ``OUTPUT_FILE``: location where the results of the algorithm should
      be stored
    - ``TOKEN_FILE``: access token for the vantage6 server REST api
    - ``USER_REQUESTED_DATABASE_LABELS``: comma-separated list of database
      labels that the user requested
    - ``<DB_LABEL>_DATABASE_URI``: uri of the each of the databases that
      the user requested, where ``<DB_LABEL>`` is the label of the
      database given in ``USER_REQUESTED_DATABASE_LABELS``.

    The wrapper expects the input file to be a json file. Any other file
    format will result in an error.

    Parameters
    ----------
    module : str
        Python module name of the algorithm to wrap.
    log_traceback: bool
        Whether to print the full error message from algorithms or not, by
        default False. Algorithm developers should set this to False if
        the error messages may contain sensitive information. By default True.
    """
    info(f"wrapper for {module}")

    # read input from the mounted input file.
    input_file = os.environ["INPUT_FILE"]
    info(f"Reading input file {input_file}")
    input_data = load_input(input_file)

    # make the actual call to the method/function
    info("Dispatching ...")
    output = _run_algorithm_method(input_data, module, log_traceback)

    # write output from the method to mounted output file. Which will be
    # transferred back to the server by the node-instance.
    output_file = os.environ["OUTPUT_FILE"]
    info(f"Writing output to {output_file}")

    _write_output(output, output_file)


def _run_algorithm_method(input_data: dict, module: str,
                          log_traceback: bool = True) -> Any:
    """
    Load the algorithm module and call the correct method to run an algorithm.

    Parameters
    ----------
    input_data : dict
        The input data that is passed to the algorithm. This should at least
        contain the key 'method' which is the name of the method that should be
        called. Another often used key is 'master' which indicates that this
        container is a master container. Other keys depend on the algorithm.
    module : str
        The module that contains the algorithm.
    log_traceback: bool, optional
        Whether to print the full error message from algorithms or not, by
        default False. Algorithm developers should set this to False if
        the error messages may contain sensitive information. By default True.

    Returns
    -------
    Any
        The result of the algorithm.
    """
    # import algorithm module
    try:
        lib = importlib.import_module(module)
        info(f"Module '{module}' imported!")
    except ModuleNotFoundError:
        error(f"Module '{module}' can not be imported! Exiting...")
        exit(1)

    # get algorithm method and attempt to load it
    method_name = input_data["method"]
    try:
        method = getattr(lib, method_name)
    except AttributeError:
        error(f"Method '{method_name}' not found!\n")
        exit(1)

    # get the args and kwargs input for this function.
    args = input_data.get("args", [])
    kwargs = input_data.get("kwargs", {})

    # try to run the method
    try:
        result = method(*args, **kwargs)
    except Exception as e:
        error(f"Error encountered while calling {method_name}: {e}")
        if log_traceback:
            error(traceback.print_exc())
        exit(1)

    return result


def load_input(input_file: str) -> Any:
    """
    Load the input from the input file.

    Parameters
    ----------
    input_file : str
        File containing the input

    Returns
    -------
    input_data : Any
        Input data for the algorithm

    Raises
    ------
    DeserializationException
        Failed to deserialize input data
    """
    with open(input_file, "rb") as fp:
        try:
            input_data = deserialization.deserialize(fp)
        except DeserializationException:
            raise DeserializationException('Could not deserialize input')
    return input_data


def _write_output(output: Any, output_file: str) -> None:
    """
    Write output to output file using JSON serialization.

    Parameters
    ----------
    output : Any
        Output of the algorithm
    output_file : str
        Path to the output file
    """
    with open(output_file, 'wb') as fp:
        serialized = serialization.serialize(output)
        fp.write(serialized)



import json
from pathlib import Path
from unittest.mock import patch, MagicMock

import pandas as pd

from vantage6.tools import wrapper

MODULE_NAME = 'algorithm_module'
DATA = 'column1,column2\n1,2'
TOKEN = 'This is a fake token'
INPUT_PARAMETERS = {'method': 'hello_world'}
SEPARATOR = '.'
SAMPLE_DB = pd.DataFrame([[1, 2]], columns=['column1', 'column2'])

MOCK_SPARQL_ENDPOINT = 'sparql://some_triplestore'


# def test_json_input_without_format_raises_deserializationexception(tmp_path):
#     """
#     It should only be possible to provide json input if it is preceded by the
#     string "json." in unicode. Otherwise a `DeserializationException` should
#     be thrown.
#     """
#     input_file = tmp_path / 'input.json'

#     with input_file.open('wb') as f:
#         f.write(json.dumps(INPUT_PARAMETERS).encode())

#     with raises(DeserializationException):
#         run_docker_wrapper_with_echo_db(input_file, tmp_path)


# def test_json_input_with_format_succeeds(tmp_path):
#     input_file = tmp_path / 'input.txt'

#     with input_file.open('wb') as f:
#         f.write(json.dumps(INPUT_PARAMETERS).encode())

#     output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)
#     assert file_echoes_db(output_file)


# def test_wrapper_serializes_json_output(tmp_path):
#     input_parameters = {'method': 'hello_world', 'output_format': JSON_FORMAT}
#     input_file = create_pickle_input(tmp_path, input_parameters)

#     output_file = run_docker_wrapper_with_echo_db(input_file, tmp_path)

#     with output_file.open('rb') as f:
#         # Check whether the data is preceded by json format string
#         assert f.read(len(JSON_FORMAT) + 1).decode() == f'{JSON_FORMAT}.'

#         # Since the echo_db algorithm was triggered, output will be table that
#         # can be read by pandas.
#         result = pd.read_json(f.read())
#         pd.testing.assert_frame_equal(SAMPLE_DB, result)


def run_docker_wrapper_with_echo_db(input_file, tmp_path):
    """
    Run the `echo_db` testing algorithm through the wrapper code. The wrapper
    communicates through files whose locations are stored in the `INPUT_FILE`,
    `TOKEN_FILE` and `DATABASE_URI` environment variables. The output of the
    algorithm is stored in the location specified in environment variable
    `OUTPUT_FILE`.

    :param input_file: input arguments to the wrapper and algorithm
    :param tmp_path: temporary path to store additional files.
    :return:
    """
    db_file = tmp_path / 'db_file.csv'
    token_file = tmp_path / 'token.txt'
    output_file = tmp_path / 'output_file.pkl'
    db_file.write_text(DATA)
    token_file.write_text(TOKEN)
    with patch('vantage6.tools.docker_wrapper.os') as mock_os:
        mock_os.environ = {
            'INPUT_FILE': input_file,
            'TOKEN_FILE': token_file,
            'OUTPUT_FILE': output_file,
            'DATABASE_URI': db_file
        }

        wrapper.docker_wrapper(MODULE_NAME)
    return output_file


@patch('vantage6.tools.wrap._run_algorithm_method')
@patch('vantage6.tools.docker_wrapper.os')
@patch('vantage6.tools.docker_wrapper.SPARQLWrapper')
def test_sparql_docker_wrapper_passes_dataframe(
    SPARQLWrapper: MagicMock, os: MagicMock, _run_algorithm_method: MagicMock,
    tmp_path: Path
):
    input_file = tmp_path / 'input_file.pkl'
    token_file = tmp_path / 'token.txt'
    output_file = tmp_path / 'output.pkl'

    environ = {'INPUT_FILE': str(input_file),
               'TOKEN_FILE': str(token_file),
               'DATABASE_URI': MOCK_SPARQL_ENDPOINT,
               'OUTPUT_FILE': str(output_file)}

    os.environ = environ

    input_args = {'query': 'select *'}

    with input_file.open('wb') as f:
        json.dumps(input_args, f)

    with token_file.open('w') as f:
        f.write(TOKEN)

    _run_algorithm_method.return_value = pd.DataFrame()
    SPARQLWrapper.return_value.query.return_value.convert.return_value = \
        DATA.encode()

    wrapper.sparql_wrapper(MODULE_NAME)

    _run_algorithm_method.assert_called_once()

    target_df = pd.DataFrame([[1, 2]], columns=['column1', 'column2'])
    pd.testing.assert_frame_equal(target_df,
                                  _run_algorithm_method.call_args[0][0])



import os
import pandas as pd

from functools import wraps
from typing import Any

from vantage6.client.algorithm_client import AlgorithmClient
from vantage6.tools.mock_client import MockAlgorithmClient
from vantage6.tools.wrap import load_input
from vantage6.tools.util import info, error, warn
from vantage6.tools.wrappers import select_wrapper


def algorithm_client(func: callable) -> callable:
    """
    Decorator that adds an algorithm client object to a function

    By adding @algorithm_client to a function, the ``algorithm_client``
    argument will be added to the front of the argument list. This client can
    be used to communicate with the server.

    There is one reserved argument `mock_client` in the function to be
    decorated. If this argument is provided, the decorator will add this
    MockAlgorithmClient to the front of the argument list instead of the
    regular AlgorithmClient.

    Parameters
    ----------
    func : callable
        Function to decorate

    Returns
    -------
    callable
        Decorated function
    """
    def wrap_function(mock_client: MockAlgorithmClient = None,
                      *args, **kwargs) -> callable:
        """
        Wrap the function with the client object

        Parameters
        ----------
        mock_client : MockAlgorithmClient
            Mock client to use instead of the regular client
        """
        if mock_client is not None:
            return func(mock_client, *args, **kwargs)
        # read server address from the environment
        host = os.environ["HOST"]
        port = os.environ["PORT"]
        api_path = os.environ["API_PATH"]

        # read token from the environment
        token_file = os.environ["TOKEN_FILE"]
        info("Reading token")
        with open(token_file) as fp:
            token = fp.read().strip()

        client = AlgorithmClient(token=token, host=host, port=port,
                                 path=api_path)
        return func(client, *args, **kwargs)
    return wrap_function


def data(number_of_databases: int = 1) -> callable:
    """
    Decorator that adds algorithm data to a function

    By adding @data to a function, one or several pandas dataframes will be
    added to the front of the argument list. This data will be read from the
    databases that the user who creates the task provides.

    Note that the user should provide exactly as many databases as the
    decorated function requires when they create the task.

    There is one reserved argument `mock_data` in the function to be
    decorated. If this argument is provided, the decorator will add this
    mocked data to the front of the argument list, instead of reading in the
    data from the databases.

    Parameters
    ----------
    number_of_databases: int
        Number of data sources to load. These will be loaded in order by which
        the user provided them. Default is 1.

    Returns
    -------
    callable
        Decorated function

    Examples
    --------
    >>> @data(number_of_databases=2)
    >>> def my_algorithm(first_df: pd.DataFrame, second_df: pd.DataFrame,
    >>>                  <other arguments>):
    >>>     pass
    """
    def protection_decorator(func: callable, *args, **kwargs) -> callable:
        @wraps(func)
        def decorator(mock_data: list[pd.DataFrame] = None, *args,
                      **kwargs) -> callable:
            """
            Wrap the function with the data

            Parameters
            ----------
            mock_data : list[pd.DataFrame]
                Mock data to use instead of the regular data
            """
            if mock_data is not None:
                return func(*mock_data, *args, **kwargs)
            # query to execute on the database
            input_file = os.environ["INPUT_FILE"]
            info(f"Reading input file {input_file}")
            input_data = load_input(input_file)

            # read the labels that the user requested, which is a comma
            # separated list of labels.
            labels = os.environ["USER_REQUESTED_DATABASE_LABELS"]
            labels = labels.split(',')

            # check if user provided enough databases
            if len(labels) < number_of_databases:
                error(f"User provided {len(labels)} databases, but algorithm "
                      f"requires {number_of_databases} databases. Exiting...")
                exit(1)
            elif len(labels) > number_of_databases:
                warn(f"User provided {len(labels)} databases, but algorithm "
                     f"requires {number_of_databases} databases. Using the "
                     f"first {number_of_databases} databases.")

            for i in range(number_of_databases):
                label = labels[i]
                data_ = _get_data_from_label(label, input_data)
                args = (data_, *args)

            return func(*args, **kwargs)
        return decorator
    return protection_decorator


def _get_data_from_label(label: str, input_data: Any) -> pd.DataFrame:
    """
    Load data from a database based on the label

    Parameters
    ----------
    label : str
        Label of the database to load
    input_data : Any
        Input data from the input file

    Returns
    -------
    pd.DataFrame
        Data from the database
    """
    # Load the input data from the input file - this may e.g. include the
    database_uri = os.environ[f"{label.upper()}_DATABASE_URI"]
    info(f"Using '{database_uri}' with label '{label}' as database")

    # Get the database type from the environment variable, this variable is
    # set by the vantage6 node based on its configuration file.
    database_type = os.environ.get(
        f"{label.upper()}_DATABASE_TYPE", "csv").lower()

    # Create the correct wrapper based on the database type, note that the
    # multi database wrapper is not available.
    wrapper = select_wrapper(database_type)
    if wrapper is None:
        error(f"Unknown database type '{database_type}' for database with "
              f"label '{label}'. Please check the node configuration.")
        exit(1)

    # Load the data from the database
    return wrapper.load_data(database_uri, input_data)
